/***************************************************************************
                          math.c  -  the math backend.
                             -------------------
    begin                : Wed Oct 11 2000
    copyright            : (C) 2000 by Brian Wagener
    email                : CptanPanic@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "stack.h"
#include "calc.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>

char DEGREE  = 0;
char HEX = 0;
char OCT = 0;
char INT = 0;
char VERBOSE = 0;

double calculate(char *str){
	struct stack infix, postfix;
	double sol = 0;

	infix = strtoin (str);
	postfix = intopost( &infix );
	sol = calpost (&postfix);
	stackdestroy(&infix);
	stackdestroy(&postfix);
	return sol;	
}

struct stack strtoin(char *str){
	char *ptr = str;
	item dat  = {0,0};
  struct stack infix;

	stackinit(&infix);	
	
	dat.op = LPAREN;
	dat.num = 0;
	push( &infix, dat);
	dat.op = 0;
	dat.num = 0;
	
	while(*ptr  ){	
		// Found a number
		if( isdigit(*ptr) || *ptr == '.' ){
			dat.num = strtod( ptr, &ptr);
			// printf("%g\n",dat.num);
		}else {// Found an op
			dat.op = strtoop( ptr, &ptr);
		}
		// Push item onto stack
		push(&infix, dat);
		dat.op = 0;
		dat.num = 0;
	}
	
	// Push a final RPAREN onto infix stack.
	dat.op = RPAREN;
	dat.num = 0;
	push( &infix, dat);
	
	return infix;
}

double factorial(double x){
	int i = 2;
	int j = floor(x);
	x = 1;
	for(  ; i <= j; i++){
		x *= i;
	}
	return x;
}	

struct stack intopost(struct stack *infix){
	struct stack postfix;
	struct stack temp;
	struct node *curr;
	item dat = {0,0};
	
	stackinit(&postfix);
	stackinit(&temp);
	
	curr = infix->head->next;
	
	while( curr ){
		if( curr->data.op == RPAREN ){
			/* Pop operators from the top of the stack and insert them in postfix
			until a left paranthesis is at the top of the stack. */
			
			 dat = pop(&temp);
			while( dat.op !=  LPAREN){
				push(&postfix,  dat );
				dat = pop(&temp);
			}
			// Discard left parens.
			if( dat.op == -1 ) //We must be missing a left paren, and we have an error.
				printf("ERROR: Left Parens expected.\n");
		} else if ( curr->data.op == LPAREN){
			// If the curent character in infix is a left parenthesis, push it onto the stack.
			push( &temp, curr->data );
					
		} else if ( curr->data.op != 0){ // We have an operator.
			/*Pop operators (if there are any) at the top of the stack (temp)
			   while they have equal or higher precedence than the current operator,
			   and insert the popped operators in postfix. */
			dat = pop(&temp);
			while( dat.op != -1 && ( curr->data.op - dat.op <= 0 ) && dat.op != LPAREN
				&& ( curr->data.op != POWER || dat.op != POWER ) ){
				// This is to deal with fact that power is right associative.
				push(&postfix,dat);
				dat = pop(&temp);
			}
			if( dat.op != -1 || dat.op == LPAREN ) push(&temp, dat);
			// Put it back since we will always pop off one to many. -1 means none left.
			push(&temp, curr->data ); // Push current op onto temp stack.
			
		} else if ( curr->data.op == 0 ){ // If we get to this point it is a digit(double)
			push( &postfix,curr->data);
		} else {
			printf("An unknown operator was found\n");
		}
		curr = curr->next;
	}	
	stackdestroy(&temp);	
	return postfix;
}

double calpost(struct stack *stk){
	struct stack cal;
	item dat;
	
	stackinit(&cal);
	
	while( stk->head->next != 0){
		double num1, num2;
		dat = pop_front(stk);
				
		if ( dat.op == PLUS){	
			dat.op = 0;
			dat.num = pop(&cal).num + pop(&cal).num ;
			push(&cal, dat);		
		} else if ( dat.op == MINUS){
			dat.op = 0;
			num1 = pop(&cal).num;
			num2 = pop(&cal).num;
			dat.num = num2 - num1;
			push(&cal, dat);
		} else if ( dat.op == DIVIDE){
			dat.op = 0;
			num1 = pop(&cal).num;
			num2 = pop(&cal).num;
			dat.num = num2 / num1 ;
			push(&cal, dat);	
		} else if ( dat.op == MULTIPLY){
			dat.op = 0;
			dat.num = pop(&cal).num * pop(&cal).num ;
			push(&cal, dat);
		} else if ( dat.op == ACOS){        // FIXME: IF out of range -1 <= x <= 1, fails.
			dat.op = 0;
			if( DEGREE )
			{
				dat.num = radtodeg( acos( pop(&cal).num ) ) ;
			} else {
       	dat.num = acos( pop(&cal).num ) ;
			}
			push(&cal, dat);
		} else if ( dat.op == COS){
			dat.op = 0;
			if( DEGREE )
			{
				dat.num =  cos( degtorad(pop(&cal).num) );
			} else {
       	dat.num =  cos( pop(&cal).num );
			}
			push(&cal, dat);
			} else if ( dat.op == COSH){
			dat.op = 0;
			dat.num =  cosh( pop(&cal).num );
			push(&cal, dat);
		} else if ( dat.op == ASIN){
			dat.op = 0;
			if( DEGREE )
			{
				dat.num = radtodeg( asin( pop(&cal).num ) ) ;
			} else {
       	dat.num = asin( pop(&cal).num ) ;
			}
			push(&cal, dat);
		} else if ( dat.op == SIN){
			dat.op = 0;
			if( DEGREE )
			{
				dat.num =  sin( degtorad(pop(&cal).num) );
			} else {
       	dat.num =  sin( pop(&cal).num );
			}
			push(&cal, dat);
		} else if ( dat.op == SINH){
			dat.op = 0;
			dat.num =  sinh( pop(&cal).num );
			push(&cal, dat);
		} else if ( dat.op == ATAN){
			dat.op = 0;
			if( DEGREE )
			{
				dat.num = radtodeg( atan( pop(&cal).num ) ) ;
			} else {
       	dat.num = atan( pop(&cal).num ) ;
			}
			push(&cal, dat);
		} else if ( dat.op == TAN){
			dat.op = 0;
			if( DEGREE )
			{
				dat.num =  tan( degtorad(pop(&cal).num) );
			} else {
       	dat.num =  tan( pop(&cal).num );
			}
			push(&cal, dat);
		} else if ( dat.op == TANH){
			dat.op = 0;
			dat.num =  tanh( pop(&cal).num );
			push(&cal, dat);
		} else if ( dat.op == EXP){
			dat.op = 0;
			dat.num =  exp( pop(&cal).num );
			push(&cal, dat);
		} else if ( dat.op == ABS){
			dat.op = 0;
			dat.num =  abs( pop(&cal).num );
			push(&cal, dat);
		} else if ( dat.op == MOD){
			dat.op = 0;
			num2 = pop(&cal).num;
			num1 = pop(&cal).num;
			dat.num =  fmod(num1,num2);
			push(&cal, dat);
		} else if ( dat.op == CEIL){
			dat.op = 0;
			dat.num =  ceil( pop(&cal).num );
			push(&cal, dat);
		} else if ( dat.op == FLOOR){
			dat.op = 0;
			dat.num =  floor( pop(&cal).num );
			push(&cal, dat);
		} else if ( dat.op == FACT){ // Not yet implemented
			dat.op = 0;
			dat.num =  factorial( pop(&cal).num );
			push(&cal, dat);	
		} else if ( dat.op == LN){
			dat.op = 0;
			dat.num =  log( pop(&cal).num );
			push(&cal, dat);
		} else if ( dat.op == LOG){
			dat.op = 0;
			dat.num =  log10( pop(&cal).num );
			push(&cal, dat);
		} else if ( dat.op == SQRT){
			dat.op = 0;
			dat.num =  sqrt( pop(&cal).num );
			push(&cal, dat);
		} else if ( dat.op == INV){
			dat.op = 0;
			dat.num =  1 / pop(&cal).num ;
			push(&cal, dat);
		} else if ( dat.op == POWER){ // Not sure it works.
			dat.op = 0;
			num2 = pop(&cal).num;
			num1 = pop(&cal).num;
			dat.num =  pow(num1,num2);
			push(&cal, dat);
		} else if ( dat.op == NEGATE	){
			dat.op = 0;
			dat.num = - pop(&cal).num;
			push(&cal, dat);
		} else { // If we get to this point, lets hope it is a double.
			push(&cal, dat);
		}
	}
	dat = pop(&cal);
	stackdestroy(&cal);
	
	return dat.num;
}


enum CALC_OP strtoop(const char * str, char **enptr){
	int index;
	char *curr = malloc(100) ;
	memset ( curr, 0, 100 );
	
	for(index = 0; str[index] != 0 ; ++index ){
		strncpy(curr,str,index + 1);
		++*enptr ;  // Increment curr string pointer
		if( strcmp( curr, "-" ) == 0 ){
			return MINUS;
		} else if ( strcmp( curr, "+" ) == 0){
			return PLUS;
		} else if ( strcmp( curr, "(" ) == 0){
			return LPAREN;
		} else if ( strcmp( curr, ")" ) == 0){
			return RPAREN;
		} else if ( strcmp( curr, "*" ) == 0){
			return MULTIPLY;
		} else if ( strcmp( curr, "/" ) == 0){
			return DIVIDE;
		} else if ( strcmp( curr, "acos" ) == 0){
			return ACOS;
		} else if ( strcmp( curr, "cos" ) == 0){
			return COS;
		} else if ( strcmp( curr, "hcos" ) == 0){
			return COSH;
		} else if ( strcmp( curr, "asin" ) == 0){
			return ASIN;
		} else if ( strcmp( curr, "sin" ) == 0){
			return SIN;
		} else if ( strcmp( curr, "hsin" ) == 0){
			return SINH;
		} else if ( strcmp( curr, "atan" ) == 0){
			return ATAN;
		} else if ( strcmp( curr, "tan" ) == 0){
			return TAN;
		} else if ( strcmp( curr, "htan" ) == 0){
			return TANH;
		} else if ( strcmp( curr, "exp" ) == 0){
			return EXP;
		} else if ( strcmp( curr, "abs" ) == 0){
			return ABS;
		} else if ( strcmp( curr, "mod" ) == 0 || strcmp( curr, "%") == 0){
			return MOD;
		} else if ( strcmp( curr, "ceil" ) == 0){
			return CEIL;
		} else if ( strcmp( curr, "floor" ) == 0){
			return FLOOR;
		} else if ( strcmp( curr, "ln" ) == 0){
			return LN;
		} else if ( strcmp( curr, "log" ) == 0){
			return LOG;
		} else if ( strcmp( curr, "sqrt" ) == 0){
			return SQRT;
		} else if ( strcmp( curr, "inv" ) == 0){
			return INV;
		} else if ( strcmp( curr, "pow" ) == 0 || strcmp( curr, "^") == 0 ){
			return POWER;
		} else if ( strcmp( curr, "neg" ) == 0){
			return NEGATE;
		} else if ( strcmp( curr, "!" ) == 0){
			return FACT;
		}
	}
	printf("mmcalc error: Unknown operator found -- %s\n",curr);
	printf("Please read documentation\n");
	exit(EXIT_FAILURE);
}

double degtorad(double x)
{
 	return ( x * M_PI / 180.0 );
}	

double radtodeg(double x)
{
 	return ( x * 180.0 / M_PI );
}

